import uuid
from collections import namedtuple
from datetime import datetime, timezone
from unittest.mock import call, patch

from django.test import override_settings
from rest_framework import status
from rest_framework.reverse import reverse

from arkindex.documents.models import Corpus, Element
from arkindex.ponos.models import GPU, Agent, Farm, State
from arkindex.process.models import (
    ActivityState,
    FeatureUsage,
    Process,
    ProcessDatasetSet,
    ProcessMode,
    Repository,
    WorkerActivity,
    WorkerVersion,
    WorkerVersionState,
)
from arkindex.project.mixins import SelectionMixin
from arkindex.project.tests import FixtureAPITestCase
from arkindex.users.models import Role


class TestCreateProcess(FixtureAPITestCase):
    """
    Test creating Workers processes with CreateProcess
    """

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.agent = Agent.objects.create(
            farm=Farm.objects.first(),
            hostname="claude",
            cpu_cores=42,
            cpu_frequency=1e15,
            ram_total=99e9,
            last_ping=datetime.now(timezone.utc),
        )
        cls.agent.gpus.create(
            id=uuid.uuid4(),
            name="claudette",
            index=2,
            ram_total=12
        )

        cls.volume = Element.objects.get(name="Volume 1")
        cls.pages = Element.objects.get_descending(cls.volume.id).filter(type__slug="page", polygon__isnull=False)
        cls.ml_class = cls.corpus.ml_classes.create(name="bretzel")

        cls.private_corpus = Corpus.objects.create(name="private")
        cls.private_corpus.elements.create(
            type=cls.private_corpus.types.create(slug="volume", folder=True),
            name="Hidden",
        )
        cls.private_ml_class = cls.private_corpus.ml_classes.create(name="chouquette")

        # Workers ProcessMode
        cls.elements_init_worker = WorkerVersion.objects.get(worker__slug="initialisation")
        cls.repo = Repository.objects.get(url="http://my_repo.fake/workers/worker")
        cls.rev_1 = cls.repo.revisions.get()
        cls.rev_2 = cls.repo.revisions.create(
            hash="2",
            message="commit message",
            author="bob",
        )
        cls.version_1 = WorkerVersion.objects.get(worker__slug="reco")
        cls.worker_1 = cls.version_1.worker
        cls.version_2 = WorkerVersion.objects.get(worker__slug="dla")
        cls.worker_2 = cls.version_2.worker
        cls.version_3 = WorkerVersion.objects.get(worker__slug="worker-gpu")
        cls.worker_3 = cls.version_3.worker

    @override_settings(
        ARKINDEX_TASKS_IMAGE="tasks:latest",
        PONOS_DEFAULT_ENV={"env": {"SOME_VAR": "somevalue"}}
    )
    def test_page_process(self):
        self.client.force_login(self.user)

        response = self.client.post(
            reverse("api:corpus-process"),
            {
                "corpus": str(self.corpus.id),
                "element_type": "page",
                "load_children": True,
            },
            format="json",
        )

        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        data = response.json()
        process = Process.objects.get(id=data["id"])
        self.assertFalse(process.tasks.exists())

        self.assertDictEqual(data, {
            "name": None,
            "id": str(process.id),
            "state": "unscheduled",
            "chunks": 1,
            "corpus": str(self.volume.corpus.id),
            "files": [],
            "mode": "workers",
            "folder_type": None,
            "element_type": "page",
            "ml_class_id": None,
            "element": None,
            "template_id": None,
            "load_children": True,
            "use_cache": False,
            "use_gpu": False,
            "activity_state": ActivityState.Disabled.value,
            "element_name_contains": None,
        })

        # The process needs to be started to produce a workflow
        process.worker_runs.create(
            version=self.version_1
        )

        response = self.client.post(
            reverse("api:process-start", kwargs={"pk": str(process.id)}),
            # The process needs a worker run or thumbnails generation to start
            {}
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        process.refresh_from_db()

        init_task = process.tasks.get(slug="initialisation")
        self.assertEqual(init_task.command, self.elements_init_worker.docker_command)
        self.assertEqual(init_task.image, self.elements_init_worker.docker_image_iid)

    def test_volume_process(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse("api:corpus-process"),
            {
                "corpus": str(self.corpus.id),
                "element": str(self.volume.id),
                "load_children": True,
                "use_cache": True,
            },
            format="json"
        )

        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        process = Process.objects.get(id=data["id"])
        self.assertFalse(process.tasks.exists())

        self.assertEqual(process.element, self.volume)
        self.assertEqual(data.pop("element")["id"], str(self.volume.id))
        self.assertDictEqual(data, {
            "name": None,
            "id": str(process.id),
            "state": "unscheduled",
            "mode": "workers",
            "chunks": 1,
            "corpus": str(self.corpus.id),
            "files": [],
            "folder_type": None,
            "element_type": None,
            "template_id": None,
            "ml_class_id": None,
            "load_children": True,
            "use_cache": True,
            "use_gpu": False,
            "activity_state": ActivityState.Disabled.value,
            "element_name_contains": None,
        })

    def test_single_page_process(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse("api:corpus-process"),
            {
                "corpus": str(self.corpus.id),
                "element": str(self.pages.first().id),
            },
            format="json"
        )

        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        process = Process.objects.get(id=data["id"])
        self.assertEqual(process.element, self.pages.first())

    def test_create_process_requires_login(self):
        response = self.client.post(reverse("api:corpus-process"), {}, format="json")
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_create_process_required_fields(self):
        self.client.force_login(self.user)
        response = self.client.post(reverse("api:corpus-process"), {}, format="json")
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(),
            {"corpus": ["This field is required."]}
        )

    @patch("arkindex.process.serializers.imports.get_max_level", return_value=Role.Contributor.value)
    def test_create_process_requires_corpus_admin(self, get_max_level_mock):
        """
        Only an admin of the target corpus can create a workers process
        """
        self.client.force_login(self.user)
        self.corpus.memberships.filter(user=self.user).update(level=Role.Contributor.value)
        response = self.client.post(
            reverse("api:corpus-process"),
            {
                "corpus": str(self.corpus.id),
                "name": "Test",
            },
            format="json"
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"corpus": ["You do not have an admin access to this corpus."]})

        self.assertEqual(get_max_level_mock.call_count, 1)
        self.assertEqual(get_max_level_mock.call_args, call(self.user, self.corpus))

    @patch("arkindex.process.serializers.imports.get_max_level", return_value=None)
    def test_create_process_non_readable_corpus(self, get_max_level_mock):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse("api:corpus-process"),
            {
                "corpus": str(self.private_corpus.id),
                "name": "Test",
            },
            format="json"
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"corpus": ["Corpus with this ID does not exist."]})

        self.assertEqual(get_max_level_mock.call_count, 1)
        self.assertEqual(get_max_level_mock.call_args, call(self.user, self.private_corpus))

    def test_create_process_name_filter(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse("api:corpus-process"),
            {
                "corpus": str(self.corpus.id),
                "element_name_contains": self.pages.first().name[2:5],
            },
            format="json"
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        process = Process.objects.get(id=data["id"])
        self.assertEqual(process.name_contains, self.pages.first().name[2:5])

    def test_create_process_fields_max_length(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse("api:corpus-process"),
            {
                "corpus": str(self.corpus.id),
                "name": "A" * 200,
                "element_name_contains": "B" * 300,
                "element_type": "C" * 100,
            },
            format="json"
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "name": ["Ensure this field has no more than 100 characters."],
            "element_name_contains": ["Ensure this field has no more than 250 characters."],
            "element_type": ["Ensure this field has no more than 50 characters."],
        })

    def test_create_process_type_filter(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse("api:corpus-process"),
            {
                "corpus": str(self.corpus.id),
                "element_type": "volume",
            },
            format="json"
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        process = Process.objects.get(id=data["id"])
        self.assertEqual(process.element_type.slug, "volume")

    def test_create_process_empty_queryset(self):
        """
        An empty QuerySet is ignored: the user can change the element filtering later,
        and running a process on no elements at all is harmless.
        """
        self.client.force_login(self.user)
        response = self.client.post(
            reverse("api:corpus-process"),
            {
                "corpus": str(self.corpus.id),
                "element_type": "volume",
                "element_name_contains": "ahahah",
            },
            format="json"
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        process = Process.objects.get(id=data["id"])
        self.assertEqual(process.corpus, self.corpus)
        self.assertEqual(process.element_type.slug, "volume")
        self.assertEqual(process.name_contains, "ahahah")
        self.assertFalse(process.list_elements().exists())

    def test_create_process_empty_selection(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse("api:corpus-process"),
            {
                "corpus": str(self.corpus.id),
                "selection": True,
            },
            format="json"
        )
        mixin = SelectionMixin()
        mixin.request = namedtuple("Request", "user")(self.user)
        self.assertEqual(mixin.get_selection(corpus_id=self.corpus.id).count(), 0)
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        data = response.json()
        self.assertIn("__all__", data)
        self.assertListEqual(
            data["__all__"],
            ["No element match those filters."]
        )

    def test_create_process_wrong_type(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse("api:corpus-process"),
            {
                "corpus": str(self.corpus.id),
                "element_type": "arachnid",
            },
            format="json"
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "element_type": ['This type does not exist in corpus "Unit Tests"']
        })

    def test_create_process_mixed_element_filters(self):
        """
        Element parameter can be set with filtering parameters
        """
        self.client.force_login(self.user)
        response = self.client.post(
            reverse("api:corpus-process"),
            {
                "corpus": str(self.corpus.id),
                "element": str(self.pages.first().id),
                "type": "page",
            },
            format="json"
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

    def test_ml_class_filter(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(9):
            response = self.client.post(
                reverse("api:corpus-process"),
                {
                    "corpus": str(self.corpus.id),
                    "ml_class_id": str(self.ml_class.id),
                },
                format="json"
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        process = self.corpus.processes.get(mode=ProcessMode.Workers, ml_class=self.ml_class)
        self.assertDictEqual(response.json(), {
            "name": None,
            "id": str(process.id),
            "state": "unscheduled",
            "chunks": 1,
            "corpus": str(self.corpus.id),
            "files": [],
            "mode": "workers",
            "folder_type": None,
            "element_type": None,
            "element": None,
            "template_id": None,
            "load_children": False,
            "use_cache": False,
            "use_gpu": False,
            "activity_state": ActivityState.Disabled.value,
            "element_name_contains": None,
            "ml_class_id": str(self.ml_class.id),
        })

    def test_create_process_non_folder_no_zone(self):
        """
        An element that is not of a folder type and has no image or polygon
        should still be usable to create a process
        """
        self.client.force_login(self.user)
        element = self.corpus.elements.create(type=self.pages.first().type, name="Kill me please")
        with self.assertNumQueries(10):
            response = self.client.post(
                reverse("api:corpus-process"),
                {
                    "corpus": str(self.corpus.id),
                    "element": str(element.id),
                },
                format="json"
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        process = Process.objects.get(element=element, mode=ProcessMode.Workers)
        self.assertDictEqual(response.json(), {
            "name": None,
            "id": str(process.id),
            "state": "unscheduled",
            "chunks": 1,
            "corpus": str(self.volume.corpus.id),
            "files": [],
            "mode": "workers",
            "folder_type": None,
            "element_type": None,
            "element": {
                "id": str(element.id),
                "name": "Kill me please",
                "type": "page",
                "zone": None,
                "corpus": {
                    "id": str(self.corpus.id),
                    "name": "Unit Tests",
                    "public": True,
                },
                "rotation_angle": 0,
                "mirrored": False,
                "thumbnail_url": None,
            },
            "template_id": None,
            "load_children": False,
            "use_cache": False,
            "use_gpu": False,
            "activity_state": ActivityState.Disabled.value,
            "element_name_contains": None,
            "ml_class_id": None,
        })

    def test_create_process_mixed_selection_filters(self):
        """
        Element and selection fields cannot be set together
        """
        self.client.force_login(self.user)
        response = self.client.post(
            reverse("api:corpus-process"),
            {
                "corpus": str(self.corpus.id),
                "element": str(self.volume.id),
                "selection": True,
            },
            format="json"
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(),
            {
                "element": ["This field cannot be set with the selection field."],
                "selection": ["This field cannot be set with the element field."]
            }
        )

    def test_create_process_selection_filter(self):
        self.client.force_login(self.user)
        page = self.pages.first()
        self.client.post(
            reverse("api:elements-selection"),
            data={"ids": [page.id]},
            format="json"
        )
        response = self.client.post(
            reverse("api:corpus-process"),
            {
                "corpus": str(self.corpus.id),
                "selection": True,
            },
            format="json"
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        process = Process.objects.get(id=data["id"])
        self.assertFalse(process.tasks.exists())
        self.assertEqual(process.elements.get(), page)

        # The process needs to be started to produce a workflow
        process.worker_runs.create(
            version=self.version_1
        )

        response = self.client.post(
            reverse("api:process-start", kwargs={"pk": str(process.id)}),
            {}
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        process.refresh_from_db()

        init_task = process.tasks.get(slug="initialisation")
        self.assertEqual(init_task.command, self.elements_init_worker.docker_command)

        response = self.client.get(reverse("api:process-elements-list", kwargs={"pk": process.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(len(data["results"]), 1)
        self.assertEqual(data["results"][0]["id"], str(page.id))

    @override_settings(ARKINDEX_FEATURES={"selection": False})
    def test_create_process_selection_no_selection(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse("api:corpus-process"),
            {
                "corpus": str(self.corpus.id),
                "selection": True,
            },
            format="json"
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"selection": ["Selection is not available on this instance."]})

    def test_create_process_process_name(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse("api:corpus-process"),
            {
                "corpus": str(self.corpus.id),
                "name": "blah",
            },
            format="json"
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        process = Process.objects.get(id=data["id"])
        self.assertEqual(process.name, "blah")

    @override_settings(
        ARKINDEX_TASKS_IMAGE="registry.teklia.com/tasks",
        PONOS_DEFAULT_ENV={}
    )
    @patch("arkindex.project.triggers.process_tasks.initialize_activity.delay")
    def test_workers_multiple_worker_runs(self, activities_delay_mock):
        process_2 = self.corpus.processes.create(
            creator=self.user,
            mode=ProcessMode.Workers,
            load_children=True,
        )
        init_run = process_2.worker_runs.create(
            version=self.elements_init_worker,
            parents=[]
        )
        run_1 = process_2.worker_runs.create(
            version=self.version_1,
            parents=[init_run.id],
        )
        run_2 = process_2.worker_runs.create(
            version=self.version_2,
            parents=[run_1.id],
        )

        self.assertFalse(process_2.tasks.exists())
        self.assertFalse(self.corpus.worker_versions.exists())

        self.client.force_login(self.user)
        with self.assertNumQueries(16):
            response = self.client.post(
                reverse("api:process-start", kwargs={"pk": str(process_2.id)}),
                {"worker_activity": True},
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(response.json()["id"], str(process_2.id))

        process_2.refresh_from_db()
        self.assertEqual(process_2.state, State.Unscheduled)

        self.assertEqual(process_2.tasks.count(), 3)

        init_task = process_2.tasks.get(slug="initialisation")
        self.assertEqual(init_task.command, self.elements_init_worker.docker_command)
        self.assertEqual(init_task.image, self.elements_init_worker.docker_image_iid)

        reco_task = process_2.tasks.get(slug=run_1.task_slug)
        self.assertEqual(reco_task.command, None)
        self.assertEqual(reco_task.image, self.version_1.docker_image_iid)
        self.assertEqual(reco_task.shm_size, None)
        self.assertEqual(list(reco_task.parents.all()), [init_task])
        self.assertEqual(reco_task.env, {
            "TASK_ELEMENTS": "/data/initialisation/elements.json",
            "ARKINDEX_WORKER_RUN_ID": str(run_1.id),
            "ARKINDEX_PROCESS_ID": str(process_2.id),
            "ARKINDEX_CORPUS_ID": str(self.corpus.id),
            "ARKINDEX_TASK_TOKEN": reco_task.token,
        })

        dla_task = process_2.tasks.get(slug=run_2.task_slug)
        self.assertEqual(dla_task.command, None)
        self.assertEqual(dla_task.image, self.version_2.docker_image_iid)
        self.assertEqual(dla_task.shm_size, None)
        self.assertEqual(list(dla_task.parents.all()), [reco_task])
        self.assertEqual(dla_task.env, {
            "TASK_ELEMENTS": "/data/initialisation/elements.json",
            "ARKINDEX_WORKER_RUN_ID": str(run_2.id),
            "ARKINDEX_PROCESS_ID": str(process_2.id),
            "ARKINDEX_CORPUS_ID": str(self.corpus.id),
            "ARKINDEX_TASK_TOKEN": dla_task.token,
        })

        # Assert two different random token have been set on the tasks
        self.assertEqual(len(reco_task.token), 44)
        self.assertEqual(len(dla_task.token), 44)
        self.assertNotEqual(reco_task.token, dla_task.token)

        # Check worker activities has been created for concerned elements on both runs
        self.assertEqual(activities_delay_mock.call_count, 1)
        self.assertEqual(activities_delay_mock.call_args, call(process_2))
        self.assertEqual(process_2.activity_state, ActivityState.Pending)

        # Manually trigger activity creations
        from arkindex.process.tasks import initialize_activity
        with self.assertNumQueries(8):
            initialize_activity(process_2)
        process_2.refresh_from_db()
        self.assertEqual(process_2.activity_state, ActivityState.Ready)
        elements_ids = self.corpus.elements.values_list("id", flat=True)
        self.assertCountEqual(
            elements_ids,
            WorkerActivity.objects.filter(worker_version=self.version_1).values_list("element_id", flat=True)
        )
        self.assertCountEqual(
            elements_ids,
            WorkerActivity.objects.filter(worker_version=self.version_2).values_list("element_id", flat=True)
        )

        # Check that the corpus worker version cache has been updated
        self.assertCountEqual(self.corpus.worker_versions.all(), [self.version_1, self.version_2])

    @override_settings(
        ARKINDEX_TASKS_IMAGE="registry.teklia.com/tasks",
        PONOS_DEFAULT_ENV={}
    )
    @patch("arkindex.ponos.models.base64.encodebytes")
    def test_create_process_use_gpu_option(self, token_mock):
        """
        A process with the `use_gpu` parameter enables the `requires_gpu` attribute of tasks that need one
        """
        token_mock.side_effect = [b"12345", b"67891"]
        process_2 = self.corpus.processes.create(creator=self.user, mode=ProcessMode.Workers)
        init_run = process_2.worker_runs.create(
            version=self.elements_init_worker,
            parents=[]
        )
        run = process_2.worker_runs.create(
            version=self.version_3,
            parents=[init_run.id],
        )
        self.client.force_login(self.user)

        with self.assertNumQueries(16):
            response = self.client.post(
                reverse("api:process-start", kwargs={"pk": str(process_2.id)}),
                {"use_gpu": True},
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        self.assertEqual(response.json()["use_gpu"], True)
        process_2.refresh_from_db()
        self.assertEqual(process_2.state, State.Unscheduled)

        init_task = process_2.tasks.get(slug="initialisation")
        self.assertEqual(init_task.command, self.elements_init_worker.docker_command)
        self.assertEqual(init_task.image, self.elements_init_worker.docker_image_iid)

        worker_task = process_2.tasks.get(slug=run.task_slug)
        self.assertEqual(worker_task.command, None)
        self.assertEqual(worker_task.image, self.version_3.docker_image_iid)
        self.assertEqual(worker_task.shm_size, None)
        self.assertEqual(worker_task.env, {
            "TASK_ELEMENTS": "/data/initialisation/elements.json",
            "ARKINDEX_CORPUS_ID": str(self.corpus.id),
            "ARKINDEX_PROCESS_ID": str(process_2.id),
            "ARKINDEX_WORKER_RUN_ID": str(run.id),
            "ARKINDEX_TASK_TOKEN": "67891"
        })
        self.assertEqual(worker_task.requires_gpu, True)
        self.assertEqual(len(worker_task.parents.all()), 1)
        self.assertEqual(worker_task.parents.first(), init_task)

    @override_settings(
        ARKINDEX_TASKS_IMAGE="registry.teklia.com/tasks",
        PONOS_DEFAULT_ENV={}
    )
    @patch("arkindex.ponos.models.base64.encodebytes")
    def test_create_process_use_gpu_option_no_available_gpus(self, token_mock):
        """
        If there are no avilables Agents with GPU, then requires_gpu is not sent if the worker
        version only supports and not requires it
        """
        self.agent.gpus.all().delete()
        token_mock.side_effect = [b"12345", b"67891", b"54321", b"19876"]

        for feature_usage, requires_gpu, task_token in [
            (FeatureUsage.Supported, False, "67891"),
            (FeatureUsage.Required, True, "19876")
        ]:
            with self.subTest(feature_usage=feature_usage, requires_gpu=requires_gpu):

                process = self.corpus.processes.create(
                    creator=self.user,
                    mode=ProcessMode.Workers,
                    farm=Farm.objects.first(),
                )
                run = process.worker_runs.create(
                    version=self.version_3,
                    parents=[],
                )
                process.use_gpu = True
                self.assertEqual(GPU.objects.count(), 0)
                self.version_3.gpu_usage = feature_usage
                self.version_3.save()
                process.run()

                init_task = process.tasks.get(slug="initialisation")
                self.assertEqual(init_task.command, self.elements_init_worker.docker_command)
                self.assertEqual(init_task.image, self.elements_init_worker.docker_image_iid)

                worker_task = process.tasks.get(slug=run.task_slug)
                self.assertEqual(worker_task.command, None)
                self.assertEqual(worker_task.image, "registry.somewhere.com/something:latest")
                self.assertEqual(worker_task.shm_size, None)
                self.assertEqual(worker_task.env, {
                    "TASK_ELEMENTS": "/data/initialisation/elements.json",
                    "ARKINDEX_CORPUS_ID": str(self.corpus.id),
                    "ARKINDEX_PROCESS_ID": str(process.id),
                    "ARKINDEX_WORKER_RUN_ID": str(run.id),
                    "ARKINDEX_TASK_TOKEN": task_token
                })
                self.assertEqual(worker_task.requires_gpu, requires_gpu)
                self.assertEqual(len(worker_task.parents.all()), 1)
                self.assertEqual(worker_task.parents.first(), init_task)

    def test_retry_keeps_requires_gpu(self):
        """
        When a process is retried, the newly created tasks keep the same requires_gpu values
        """
        process = self.corpus.processes.create(
            creator=self.user,
            mode=ProcessMode.Workers,
            farm=Farm.objects.first(),
        )
        process.worker_runs.create(
            version=self.version_3,
            parents=[],
        )
        process.use_gpu = True
        process.run()
        self.assertEqual(process.state, State.Unscheduled)
        self.client.force_login(self.user)

        init_task, worker_task = process.tasks.filter(run=0)
        self.assertEqual(init_task.requires_gpu, False)
        self.assertEqual(worker_task.requires_gpu, True)

        process.tasks.update(state=State.Stopped)

        response = self.client.post(reverse("api:process-retry", kwargs={"pk": str(process.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        init_task, worker_task = process.tasks.filter(run=1)
        self.assertEqual(init_task.requires_gpu, False)
        self.assertEqual(worker_task.requires_gpu, True)

    def test_use_gpu_unsupported(self):
        process = self.corpus.processes.create(creator=self.user, mode=ProcessMode.Workers)
        process.worker_runs.create(
            version=self.version_2,
            parents=[],
        )
        process.use_gpu = True
        process.save()
        self.client.force_login(self.user)
        with self.assertNumQueries(5):
            response = self.client.post(
                reverse("api:process-start", kwargs={"pk": str(process.id)}),
                {"use_gpu": "true"}
            )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(
            response.json(),
            {"use_gpu": ["The process is configured to use a GPU, but does not include any workers that support GPU usage."]}
        )

    def test_use_gpu_some_unsupported(self):
        process = self.corpus.processes.create(creator=self.user, mode=ProcessMode.Workers)
        init_run = process.worker_runs.create(
            version=self.elements_init_worker,
            parents=[]
        )
        process.worker_runs.create(
            version=self.version_2,
            parents=[init_run.id],
        )
        process.worker_runs.create(
            version=self.version_3,
            parents=[init_run.id],
        )
        process.use_gpu = True
        process.save()
        self.client.force_login(self.user)
        with self.assertNumQueries(16):
            response = self.client.post(
                reverse("api:process-start", kwargs={"pk": str(process.id)}),
                {"use_gpu": "true"}
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        process.refresh_from_db()
        init_task, worker_2_task, worker_3_task = process.tasks.filter(run=0)
        self.assertEqual(init_task.requires_gpu, False)
        self.assertEqual(worker_2_task.requires_gpu, False)
        self.assertEqual(worker_3_task.requires_gpu, True)

    def test_dataset_gpu_required(self):
        self.client.force_login(self.user)
        process = self.corpus.processes.create(creator=self.user, mode=ProcessMode.Dataset)
        dataset = self.corpus.datasets.first()
        test_set = dataset.sets.get(name="test")
        ProcessDatasetSet.objects.create(process=process, set=test_set)
        process.versions.set([self.version_2, self.version_3])

        with self.assertNumQueries(9):
            response = self.client.post(reverse("api:process-start", kwargs={"pk": str(process.id)}))
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(response.json(), [
            "Some worker versions require a GPU and the `use_gpu` option is disabled.",
        ])

    def test_workers_gpu_required(self):
        self.client.force_login(self.user)
        process = self.corpus.processes.create(creator=self.user, mode=ProcessMode.Workers)
        process.versions.set([self.version_2, self.version_3])

        with self.assertNumQueries(9):
            response = self.client.post(reverse("api:process-start", kwargs={"pk": str(process.id)}))
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(response.json(), [
            "Some worker versions require a GPU and the `use_gpu` option is disabled.",
        ])

    def test_dataset_archived(self):
        self.client.force_login(self.user)
        self.worker_1.archived = datetime.now(timezone.utc)
        self.worker_1.save()
        process = self.corpus.processes.create(creator=self.user, mode=ProcessMode.Dataset)
        dataset = self.corpus.datasets.first()
        test_set = dataset.sets.get(name="test")
        ProcessDatasetSet.objects.create(process=process, set=test_set)
        process.versions.add(self.version_1)

        with self.assertNumQueries(9):
            response = self.client.post(reverse("api:process-start", kwargs={"pk": str(process.id)}))
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(response.json(), [
            "Some worker versions are on archived workers and cannot be executed.",
        ])

    def test_workers_archived(self):
        self.client.force_login(self.user)
        self.worker_1.archived = datetime.now(timezone.utc)
        self.worker_1.save()

        process = self.corpus.processes.create(creator=self.user, mode=ProcessMode.Workers)
        process.versions.add(self.version_1)

        with self.assertNumQueries(9):
            response = self.client.post(reverse("api:process-start", kwargs={"pk": str(process.id)}))
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(response.json(), [
            "Some worker versions are on archived workers and cannot be executed.",
        ])

    def test_start_no_revision(self):
        """
        A worker version without a revision but with a Docker image should be usable in a process
        """
        custom_version = WorkerVersion.objects.get(worker__slug="custom")
        self.assertIsNone(custom_version.revision_id)
        custom_version.docker_image_iid = "alpine:latest"
        custom_version.state = WorkerVersionState.Available
        custom_version.save()

        self.client.force_login(self.user)
        process = self.corpus.processes.create(creator=self.user, mode=ProcessMode.Workers)
        init_run = process.worker_runs.create(
            version=self.elements_init_worker,
            parents=[]
        )
        process.worker_runs.create(
            version=custom_version,
            parents=[init_run.id]
        )

        with self.assertNumQueries(16):
            response = self.client.post(reverse("api:process-start", kwargs={"pk": str(process.id)}))
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        self.assertDictEqual(response.json(), {
            "name": None,
            "id": str(process.id),
            "state": "unscheduled",
            "chunks": 1,
            "corpus": str(self.corpus.id),
            "files": [],
            "mode": "workers",
            "folder_type": None,
            "element_type": None,
            "element": None,
            "template_id": None,
            "load_children": False,
            "use_cache": False,
            "use_gpu": False,
            "activity_state": ActivityState.Disabled.value,
            "element_name_contains": None,
            "ml_class_id": None,
        })

        task = process.tasks.get(slug__startswith="custom_")
        self.assertEqual(task.image, "alpine:latest")

    def test_create_process_valid_mode(self):
        self.client.force_login(self.user)
        cases = ["dataset", "workers", None]
        for mode in cases:
            data = {
                "name": "Test Process",
                "corpus": str(self.corpus.id),
            }
            if mode:
                data["mode"] = mode
            with (self.subTest(mode=mode), self.assertNumQueries(8)):
                response = self.client.post(
                    reverse("api:corpus-process"),
                    data,
                    format="json"
                )
                self.assertEqual(response.status_code, status.HTTP_201_CREATED)

            process = self.corpus.processes.get(id=response.json()["id"])
            self.assertEqual(process.mode, ProcessMode(mode) if mode else ProcessMode.Workers)
            self.assertDictEqual(response.json(), {
                "name": "Test Process",
                "id": str(process.id),
                "state": "unscheduled",
                "chunks": 1,
                "corpus": str(self.corpus.id),
                "files": [],
                "mode": mode if mode else "workers",
                "folder_type": None,
                "element_type": None,
                "element": None,
                "template_id": None,
                "load_children": False,
                "use_cache": False,
                "use_gpu": False,
                "activity_state": ActivityState.Disabled.value,
                "element_name_contains": None,
                "ml_class_id": None,
            })

    def test_create_process_dataset_mode_no_selection(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.post(
                reverse("api:corpus-process"),
                {
                    "name": "Test Process",
                    "corpus": str(self.corpus.id),
                    "mode": "dataset",
                    "selection": True
                },
                format="json"
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "mode": ["This process mode cannot be set with the selection field."],
            "selection": ["This field cannot be set with the specified process mode."]
        })

    def test_create_process_invalid_mode(self):
        self.client.force_login(self.user)
        cases = [item.value for item in ProcessMode if item.value not in ["workers", "dataset"]]
        for mode in cases:
            with (self.subTest(mode=mode), self.assertNumQueries(6)):
                response = self.client.post(
                    reverse("api:corpus-process"),
                    {
                        "name": "Test Process",
                        "corpus": str(self.corpus.id),
                        "mode": mode,
                    },
                    format="json"
                )
                self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
                self.assertDictEqual(response.json(), {"mode": [f'"{mode}" is not a valid choice.']})
